/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    IB1.java
 *    Copyright (C) 1999 Stuart Inglis,Len Trigg,Eibe Frank
 *
 */

package weka.classifiers.misc;

import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.classifiers.UpdateableClassifier;
import java.io.*;
import java.util.*;
import weka.core.*;
import Jama.CholeskyDecomposition;
import Jama.Matrix;
import java.text.NumberFormat;

/**
 * Anderson-Gaborsky Polynomial Network.  Uses a pseudo-inverse to create
 * a weight matrix to use as classifier.  This wieght matrix is then trained
 * iteratively.  For more information, see <p>
 * 
 * Anderson, Peter G. and Roger S. Gaborski (1993) "The Polynomial Method
 * Augmented by Supervised Training for Hand-Printed Character Recognition<p>
 *
 * @author Sam Danziger (Sam_Danziger@ieee.org)
 * @version $Revision: 1.0 $
 */
public class WekaPolynet extends Classifier {

  /** The Polynet Classifier **/
  private Polynet myPolynet;
  
  private Matrix x;  //  The feature matrix
  private Matrix y;  //  The class matrix
  
  private int maxEpoch;  //  The maximum number of epochs to train for
  
  /** The training instances used for classification. */
  private Instances m_Train;

  private ArrayList classLUT;
  

  /**
   * Returns a string describing classifier
   * @return a description suitable for
   * displaying in the explorer/experimenter gui
   */
  public String globalInfo() {

    return "Anderson-Gaborsky Polynomial Network.  Uses a pseudo-inverse to create" 
      + "a weight matrix to use as classifier.  This wieght matrix is then trained"
      + "iteratively.  For more information, see \n\n"
      + "Anderson, Peter G. and Roger S. Gaborski (1993) \"The Polynomial Method "
      + "Augmented by Supervised Training for Hand-Printed Character Recognition\",";
  }

  /**
   * Generates the classifier.
   *
   * @param instances set of instances serving as training data 
   * @exception Exception if the classifier has not been generated successfully
   */
  public void buildClassifier(Instances instances) throws Exception {
    
    if (instances.checkForStringAttributes()) {
      throw new UnsupportedAttributeTypeException("Cannot handle string attributes!");
    }
    // Throw away training instances with missing class
    m_Train = new Instances(instances, 0, instances.numInstances());
    m_Train.deleteWithMissingClass();

    x = new Matrix(m_Train.numInstances(),m_Train.numAttributes()-1);
    y = new Matrix(m_Train.numInstances(),m_Train.numClasses());
    System.out.println("numClasses = "+m_Train.numClasses());
    
    makeLUT();
    Enumeration enum = m_Train.enumerateInstances();
    int ctr = 0;
    while (enum.hasMoreElements()) {
      addNextInstance((Instance) enum.nextElement(),ctr);
      ctr++;
    }
    
    x.print(NumberFormat.getInstance(), 4);
    y.print(NumberFormat.getInstance(), 4);
    
    maxEpoch = 40;
    myPolynet = new Polynet(maxEpoch, "no", x, y);
  }
  
  
  /**
   *Populate the class LUT
   *
   */
  private void makeLUT(){
  	LinkedHashSet lhs_classLUT = new LinkedHashSet();
  	
  	Enumeration enum = m_Train.enumerateInstances();
    while (enum.hasMoreElements()) {
      Instance inst = (Instance) enum.nextElement();
      lhs_classLUT.add(new Double(inst.classValue()));
    }
    
    classLUT = new ArrayList(lhs_classLUT);
    System.out.println("classLUT size= "+classLUT.size());
  }
  
  
  /**
   *Return the index from the class Look up table
   *
   *@param classVal
   */
  double [] lutIndex(Double classVal){
  	int idx = classLUT.indexOf(classVal);
  	
  	double [] binArray = new double[classLUT.size()];
  	
  	for (int i=0; i< classLUT.size(); i++){
  		if (i==idx){
  			binArray[i]=1;
  		} else {
  			binArray[i]=0;
  		}
  	}
  	
  	return binArray;
  }
  
  
  /**
   * Add an instance to the X and Y matrix
   *
   * @param instance the new instance
   * @param instanceNum  The index of the new instance
   */
   private void addNextInstance(Instance instance, int instanceNum){
   	//System.out.println(m_Train.numAttributes());
    for (int j = 0;j < m_Train.numAttributes()-1; j++) {
    //if ((m_Train.attribute(j).isNumeric()) && (!instance.isMissing(j))) {
      if (instance.attribute(j).isNumeric() && !instance.isMissing(j)) {
      	x.set(instanceNum,j,instance.value(j));
      } else if (!instance.isMissing(j) ) { //isNominal
      	x.set(instanceNum,j,instance.value(j)+1);
      	//System.out.println("instance.value(j)="+instance.value(j));
      }
    }
    
    Double classDouble = new Double(instance.classValue());
    int idx = classLUT.indexOf(classDouble);    
   	y.set(instanceNum,idx,1);
  }
  
  /**
   * Classifies the given test instance.
   *
   * @param instance the instance to be classified
   * @return the predicted class for the instance 
   * @exception Exception if the instance can't be classified
   */
  public double classifyInstance(Instance instance) throws Exception {
    
    if (m_Train.numInstances() == 0) {
      throw new Exception("No training instances!");
    }
    
    int numAttr = instance.numAttributes()-1; //included class as an attr
    
	Matrix newX = new Matrix(1,numAttr);
	for (int i=0; i<numAttr; i++){
		newX.set(0,i,instance.value(i));
	}
	
	int classIdx = myPolynet.classify(newX);
	Double dClass = (Double)(classLUT.get(classIdx));

	return dClass.doubleValue();
  }

  /**
   * Returns a description of this classifier.
   *
   * @return a description of this classifier as a string.
   */
  public String toString() {

    return ("Polynet classifier");
  }

  /**
   *Set the options
   *
   *-N train for N epochs, default is 20
   *
   */
  public void setOptions(String[] options) throws Exception {
  	
  	String epochString = Utils.getOption('N', options);
    if (epochString.length() != 0) {
      maxEpoch = Integer.parseInt(epochString);
    } else {
      maxEpoch = 40;
    }
  }
  

  /**
   * Main method for testing this class.
   *
   * @param argv should contain command line arguments for evaluation
   * (see Evaluation).
   */
  public static void main(String [] argv) {

    try {
      System.out.println(Evaluation.evaluateModel(new WekaPolynet(), argv));
    } catch (Exception e) {
      System.err.println(e.getMessage());
    }
  }
}




